@@ -44,6 +44,9 @@ struct _GnomeCupsPrinterDetails {
 
 	guint is_default : 1;
 	guint is_gone : 1;
+	guint is_local : 1;
+
+	guint attributes_request_id;
 
 	/* Option management */
 	guint options_invalid : 1;
@@ -56,6 +59,7 @@ struct _GnomeCupsPrinterDetails {
 	char *info;
 	char *make_and_model;
 	char *device_uri;
+	char *printer_uri;
 	char *state_message;
 	ipp_pstate_t state;
 	int job_count;
@@ -205,34 +209,22 @@ gnome_cups_printer_get_info (GnomeCupsPrinter *printer)
 	return printer->details->info;
 }
 
-static void
-update_attributes (GnomeCupsPrinter *printer)
-{
 #define MAP_INT(v,a) {if (!g_ascii_strcasecmp (attr->name, (a))) { if ((v) != attr->values[0].integer) { changed = TRUE; } (v) = attr->values[0].integer; }}
 #define MAP_STRING(v,a) {if (!g_ascii_strcasecmp (attr->name, (a))) { if (!v || strcmp (v, attr->values[0].string.text)) { g_free (v); changed = TRUE; (v) = g_strdup (attr->values[0].string.text); }}}
 
-	ipp_t *request;
-	ipp_t *response;	
+static void
+attributes_update_cb (guint id,
+		      const char *path,
+		      ipp_t *response,
+		      GError **error,
+		      gpointer cb_data)
+{
+	GnomeCupsPrinter *printer;
 	ipp_attribute_t *attr;
 	gboolean changed;
-	GError *error = NULL;
-	static const char *attributes[] = {
-	  "printer-state", "queued-job-count",
-	  "printer-location", "printer-info",
-	  "printer-state-message", "device-uri",
-	  "printer-state-reasons", "printer-info",
-	  "printer-make-and-model"
-	};
-	
-	request = gnome_cups_request_new_for_printer (IPP_GET_PRINTER_ATTRIBUTES,
-						      printer->details->printer_name);
-	gnome_cups_request_add_requested_attributes (request, 
-						     IPP_TAG_OPERATION,
-						     G_N_ELEMENTS (attributes),
-						     (char**)attributes);
 
-	response = gnome_cups_request_execute (request, "/", &error);
-	
+	printer = GNOME_CUPS_PRINTER (cb_data);
+
 	changed = FALSE;
 
 	if (!error && response) {
@@ -254,12 +246,11 @@ update_attributes (GnomeCupsPrinter *printer)
 			MAP_STRING (printer->details->state_message, "printer-state-message");
 			MAP_STRING (printer->details->info, "printer-info");
 			MAP_STRING (printer->details->make_and_model, "printer-make-and-model");
+			MAP_STRING (printer->details->printer_uri, "printer-uri-supported");
 		}
 	}
 	ippDelete (response);
-	if (error) {
-		g_error_free (error);
-	}
+	g_clear_error (error);
 
 	if (changed) {
 		g_free (printer->details->full_state);
@@ -267,8 +258,42 @@ update_attributes (GnomeCupsPrinter *printer)
 		g_signal_emit (printer, signals[ATTRIBUTES_CHANGED], 0);
 	}
 
+	printer->details->attributes_request_id = 0;
+}
+
 #undef MAP_INT
 #undef MAP_STRING
+
+
+static void
+update_attributes (GnomeCupsPrinter *printer)
+{
+	ipp_t *request;
+	static const char *attributes[] = {
+	  "printer-state", "queued-job-count",
+	  "printer-location", "printer-info",
+	  "printer-state-message", "device-uri",
+	  "printer-state-reasons", "printer-info",
+	  "printer-make-and-model", "printer-uri-supported"
+	};
+
+	if (printer->details->attributes_request_id > 0) {
+		return;
+	}
+
+	request = gnome_cups_request_new_for_printer (IPP_GET_PRINTER_ATTRIBUTES,
+						      printer->details->printer_name);
+
+	gnome_cups_request_add_requested_attributes (request, 
+						     IPP_TAG_OPERATION,
+						     G_N_ELEMENTS (attributes),
+						     (char**)attributes);
+	printer->details->attributes_request_id = 
+		gnome_cups_request_execute_async (request, NULL, "/",
+						  attributes_update_cb,
+						  g_object_ref (printer),
+						  g_object_unref);
+	
 }
 
 static char *
@@ -308,7 +333,7 @@ get_default (void)
 	cupsFreeDests (num_dests, dests);
 
 	request = gnome_cups_request_new (CUPS_GET_DEFAULT);
-	response = gnome_cups_request_execute (request, "/", &error);
+	response = gnome_cups_request_execute (request, NULL, "/", &error);
 
 	if (error) {
 		ippDelete (response);
@@ -425,7 +450,7 @@ get_printer_names (void)
 	
 	request = gnome_cups_request_new (CUPS_GET_PRINTERS);
 	
-	response = gnome_cups_request_execute (request, "/", &error);
+	response = gnome_cups_request_execute (request, NULL, "/", &error);
 
 	if (error) {
 		ippDelete (response);
@@ -501,7 +526,7 @@ update_printers_timeout (gpointer user_data)
 {
 	/* To avoid unneccessary calls during authentication, check
 	 * if a request is currently executing */
-	if (!_gnome_cups_request_is_executing ()) {
+	if (_gnome_cups_outstanding_request_count () == 0) {
 		update_printers ();
 	}
 	
@@ -512,13 +537,8 @@ static void
 set_timeout (void)
 {
 	static guint update_timeout_id = 0;
-	gboolean should_timeout;
-	
-	if (add_notifies || g_hash_table_size (printers) > 0) {
-		should_timeout = TRUE;
-	} else {
-		should_timeout = FALSE;
-	}	
+	gboolean should_timeout =
+		add_notifies || (printers != NULL && g_hash_table_size (printers) > 0);
 
 	if (should_timeout && !update_timeout_id) {
 		update_timeout_id = g_timeout_add (UPDATE_TIMEOUT, 
@@ -604,7 +624,7 @@ check_exists (const char *printer)
 
 	request = gnome_cups_request_new_for_printer (IPP_GET_PRINTER_ATTRIBUTES,
 						      printer);
-	response = gnome_cups_request_execute (request, "/", &error);
+	response = gnome_cups_request_execute (request, NULL, "/", &error);
 
 	ippDelete (response);
 
@@ -674,26 +694,11 @@ gnome_cups_printer_is_gone (GnomeCupsPrinter *printer)
 	return printer->details->is_gone;
 }
 
-static gboolean
-is_local (GnomeCupsPrinter *printer)
+gboolean
+gnome_cups_printer_get_is_local (GnomeCupsPrinter *printer)
 {
-	int i;
-	static const char *local_uri_schemes[] = {
-		"usb", "serial", "parallel", "epson", "canon", NULL
-	};
-	
-	if (!printer->details->device_uri) {
-		return FALSE;
-	}
-
-	for (i = 0; local_uri_schemes[i] != NULL; i++) {
-		if (!strncmp (printer->details->device_uri, 
-			      local_uri_schemes[i],
-			      strlen (local_uri_schemes[i]))) {
-			return TRUE;
-		}
-	}
-	return FALSE;
+	return (printer->details->device_uri != NULL) && 
+		(strcmp (printer->details->device_uri, "") != 0);
 }
 
 void
@@ -704,7 +709,7 @@ gnome_cups_printer_get_icon (GnomeCupsPrinter *printer,
 	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
 	g_return_if_fail (name != NULL);
 
-	*name = is_local (printer) ? g_strdup ("gnome-dev-printer") : g_strdup ("gnome-dev-printer-network");
+	*name = gnome_cups_printer_get_is_local (printer) ? g_strdup ("gnome-dev-printer") : g_strdup ("gnome-dev-printer-network");
 	if (emblems) {
 		*emblems = NULL;
 		if (gnome_cups_printer_get_state (printer) == IPP_PRINTER_STOPPED) {
@@ -814,7 +819,7 @@ gnome_cups_printer_set_description (GnomeCupsPrinter *printer,
 						      printer->details->printer_name);
 	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
 		      "printer-info", NULL, description);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
 }
@@ -846,7 +851,7 @@ gnome_cups_printer_set_location (GnomeCupsPrinter *printer,
 						      printer->details->printer_name);
 	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
 		      "printer-location", NULL, location);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
 }
@@ -862,7 +867,7 @@ gnome_cups_printer_pause (GnomeCupsPrinter *printer,
 	
 	request = gnome_cups_request_new_for_printer (IPP_PAUSE_PRINTER,
 						      printer->details->printer_name);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
 }
@@ -878,7 +883,7 @@ gnome_cups_printer_resume (GnomeCupsPrinter *printer,
 	
 	request = gnome_cups_request_new_for_printer (IPP_RESUME_PRINTER,
 						      printer->details->printer_name);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);	
 }
@@ -894,7 +899,7 @@ gnome_cups_printer_delete (GnomeCupsPrinter *printer,
 
 	request = gnome_cups_request_new_for_printer (CUPS_DELETE_PRINTER,
 						      printer->details->printer_name);
-	response = gnome_cups_request_execute (request, "/admin/", error);
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 }
 
@@ -1250,6 +1255,9 @@ static void
 gnome_cups_printer_finalize (GObject *object)
 {
 	GnomeCupsPrinter *printer = GNOME_CUPS_PRINTER (object);
+
+	if (printer->details->attributes_request_id > 0)
+		gnome_cups_request_cancel (printer->details->attributes_request_id > 0);
 	
 	if (printer->details->ppd_options) {
 		g_hash_table_destroy (printer->details->ppd_options);
@@ -1402,7 +1410,7 @@ gnome_cups_printer_print_file (GnomeCupsPrinter *printer,
 	GList *l;
 
 	g_return_val_if_fail (printer != NULL, FALSE);
-	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer) != NULL, FALSE);
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), FALSE);
 	g_return_val_if_fail (printer->details != NULL, FALSE);
 	g_return_val_if_fail (printer->details->printer_name != NULL, FALSE);
 
@@ -1489,3 +1497,47 @@ gnome_cups_printer_force_refresh (GnomeCupsPrinter        *printer,
 		printer->details->options_invalid = TRUE;
 	}
 }
+
+gchar *
+_gnome_cups_printer_get_host (GnomeCupsPrinter *printer)
+{
+	gchar *host = NULL;
+	if (printer->details->printer_uri) {
+		gchar *x, *y;
+
+		x = strstr (printer->details->printer_uri, "://");
+
+		if (x) {
+			x += 3;
+			y = strpbrk (x, ":/");
+			if (y)
+				host = g_strndup (x, y - x);
+			else 
+				host = g_strdup (x);
+		}
+	}
+	
+	return host;
+}
+
+char *
+gnome_cups_get_printer_uri (const char *printer_name)
+{
+	char *uri;
+	GnomeCupsPrinter *printer;
+
+	if (printer_name) {
+		printer = gnome_cups_printer_get_existing (printer_name);
+		if (printer && printer->details->printer_uri)
+			uri = g_strdup (printer->details->printer_uri);
+		else
+			uri = g_strdup_printf ("ipp://localhost/printers/%s",
+					       printer_name);
+		if (printer)
+			g_object_unref (printer);
+	} else {
+		uri = g_strdup_printf ("ipp://localhost/printers/");
+	}
+
+	return uri;
+}
